//  The extended procedures implemented in this DLL are:
//
//  XP_PROCLIST	Returns all this DLL's extended stored procedures and 
//					their usuage.
//
//  XP_ECHO		Used to demo the handling of output parameters in
//					extended procedures.
//
//
// Miscellaneous defines
//
#define  XP_NOERROR		0
#define  XP_ERROR		1


// Stored Procedure information structure.
//
typedef struct xp_info {
    DBCHAR name[MAXNAME];   	// Extended procedure name
    DBCHAR usage[MAXLEN];   		// Usage string
} XP_INFO;

// Array of Extended Stored Procedures supported by this DLL.
//
XP_INFO Xps[] = 
{
    "xp_proclist",					// Procedure name
    "usage: xp_proclist",		 	// Procedure usage string

    "xp_echo",
    "usage: xp_echo @p1=<input> , @p2=<output-param> output",

};
#define Xpcnumber sizeof(Xps) / sizeof(XP_INFO)

//
// PROCLIST
//    Returns the usage for all defined stored procedures
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    XP_NOERROR
//    XP_ERROR
//
// Side Effects:
//    Returns a result set to client
//
RETCODE xp_proclist(srvproc)
SRV_PROC *srvproc;
{
    int paramnum;
    DBCHAR colname1[MAXNAME];
    DBCHAR colname2[MAXNAME];
    int i;

    // Get number of parameters
    //
    paramnum = srv_rpcparams(srvproc);

    // Check number of parameters
    //
    if (paramnum != -1) {
        // Send error message and return
        //
        srv_sendmsg(srvproc, SRV_MSG_ERROR, XP_PROCLIST_ERROR, SRV_INFO, (DBTINYINT)0,
                    NULL, 0, 0, "Error executing extended stored procedure: Invalid Parameter",
                    SRV_NULLTERM);
	 	  // A SRV_DONE_MORE instead of a SRV_DONE_FINAL must complete the
	 	  // result set of an Extended Stored Procedure.
	 	  //
        srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_MORE), 0, 0);
	return(XP_ERROR);
    }

    sprintf(colname1, "spname");
    srv_describe(srvproc, 1, colname1, SRV_NULLTERM, SRVCHAR, MAXNAME,
                 SRVCHAR, 0, NULL);

    sprintf(colname2, "spusage");
    srv_describe(srvproc, 2, colname2, SRV_NULLTERM, SRVCHAR, MAXLEN, SRVCHAR,
                 0, NULL);

    // Return each XP handler as a row
    //
    for (i = 0; i < Xpcnumber; i++) {
        srv_setcoldata(srvproc, 1, Xps[i].name);
        srv_setcollen(srvproc, 1, strlen(Xps[i].name));

        srv_setcoldata(srvproc, 2, Xps[i].usage);
        srv_setcollen(srvproc, 2, strlen(Xps[i].usage));

        srv_sendrow(srvproc);
    }
    srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_MORE), 0, i);
    return(XP_NOERROR);
}

//
// XP_ECHO
//	Used to demo the handling of output parameters in extended procedures.
//	The first parameter should be the input parameter and the second
//	parameter should be the output.  The input parameter is placed in the
//	output parameter and returned.
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    XP_NOERROR
//    XP_ERROR
//
//

RETCODE xp_echo(srvproc)
SRV_PROC *srvproc;
{
    int paramnum;

    // Check number of parameters
    //
    if ((paramnum = srv_rpcparams(srvproc)) != 2) {
        srv_sendmsg(srvproc, SRV_MSG_ERROR, ECHO_ERROR, SRV_INFO, (DBTINYINT)0,
		    NULL, 0, 0, "Error executing extended stored procedure: Invalid # of Parameters",
                    SRV_NULLTERM);
        srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_MORE), 0, 0);
	return(XP_ERROR);
    }

	// check for param names if desired
	// check to see if param types match
	// check to see that P2 is declared as ouput 
	// etc.

    // Set output parameter to input parameter.
    //

    srv_paramset(srvproc, 2, srv_paramdata(srvproc,1), srv_paramlen(srvproc,1));

 	// A SRV_DONE_MORE instead of a SRV_DONE_FINAL must complete the
	// result set of an Extended Stored Procedure.
		
    srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);

    return(XP_NOERROR);

}


