//
//  The extended procedures implemented in this DLL are:
//
//  XP_PROCLIST	Returns all this DLL's extended stored procedures and 
//					their usuage.
//
//   xp_traceon   shows pre- handlers to trace output to console
//
//   xp_traceoff   de-installs pre- handler
//


// Miscellaneous defines
//
#define  XP_NOERROR		0
#define  XP_ERROR		1


// function prototypes for xp_trace callbacks
RETCODE trace_lang(SRV_PROC *);
RETCODE trace_rpc(SRV_PROC *);

// Stored Procedure information structure.
//
typedef struct xp_info {
    DBCHAR name[MAXNAME];   	// Extended procedure name
    DBCHAR usage[MAXLEN];   		// Usage string
} XP_INFO;

// Array of Extended Stored Procedures supported by this DLL.
//
XP_INFO Xps[] = 
{
    "xp_proclist",					// Procedure name
    "usage: xp_proclist",		 	// Procedure usage string

	"xp_traceon",
    "usage: xp_traceon [@p1=-1] , -1=all connections",

    "xp_traceoff",
    "usage: xp_traceoff [@p1=-1] , -1=all connections",
};
#define Xpcnumber sizeof(Xps) / sizeof(XP_INFO)

//
// PROCLIST
//    Returns the usage for all defined stored procedures
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    XP_NOERROR
//    XP_ERROR
//
// Side Effects:
//    Returns a result set to client
//
RETCODE xp_proclist(srvproc)
SRV_PROC *srvproc;
{
    int paramnum;
    DBCHAR colname1[MAXNAME];
    DBCHAR colname2[MAXNAME];
    int i;

    // Get number of parameters
    //
    paramnum = srv_rpcparams(srvproc);

    // Check number of parameters
    //
    if (paramnum != -1) {
        // Send error message and return
        //
        srv_sendmsg(srvproc, SRV_MSG_ERROR, XP_PROCLIST_ERROR, SRV_INFO, (DBTINYINT)0,
                    NULL, 0, 0, "Error executing extended stored procedure: Invalid Parameter",
                    SRV_NULLTERM);
	 	  // A SRV_DONE_MORE instead of a SRV_DONE_FINAL must complete the
	 	  // result set of an Extended Stored Procedure.
	 	  //
        srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_MORE), 0, 0);
	return(XP_ERROR);
    }

    sprintf(colname1, "spname");
    srv_describe(srvproc, 1, colname1, SRV_NULLTERM, SRVCHAR, MAXNAME,
                 SRVCHAR, 0, NULL);

    sprintf(colname2, "spusage");
    srv_describe(srvproc, 2, colname2, SRV_NULLTERM, SRVCHAR, MAXLEN, SRVCHAR,
                 0, NULL);

    // Return each XP handler as a row
    //
    for (i = 0; i < Xpcnumber; i++) {
        srv_setcoldata(srvproc, 1, Xps[i].name);
        srv_setcollen(srvproc, 1, strlen(Xps[i].name));

        srv_setcoldata(srvproc, 2, Xps[i].usage);
        srv_setcollen(srvproc, 2, strlen(Xps[i].usage));

        srv_sendrow(srvproc);
    }
    srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_MORE), 0, i);
    return(XP_NOERROR);
}



//
// XP_TRACEON
//	Used to demo the pre-handlers in extended procedures.
//	The first parameter should be the spid of the connection to be traced.
//	If 0, will trace the current spid.  If -1, trace all spids.
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    XP_NOERROR
//    XP_ERROR
//
//

RETCODE xp_traceon(srvproc)
SRV_PROC *srvproc;
{
    int paramnum;
	int paramtype;
	int trace_num;
	SRV_SERVER *server;
	SRV_CONFIG *config;
	char timestr[9];

    // Check number, type of parameter here
    //
    
    // 
	// Fetch param value.
	if (paramnum == 1) 
		srv_bmove(srv_paramdata(srvproc, 1), &trace_num, (DBINT) 4);
	else
		trace_num = 0;
	switch(trace_num) {
		case 0:
			// trace only this current connection
			srv_pre_handle(NULL,srvproc,SRV_LANGUAGE,trace_lang,FALSE);
			srv_pre_handle(NULL,srvproc,SRV_RPC,trace_rpc,FALSE);
			printf("%s  KPID %s  starting trace \n", _strtime(timestr),
    			srv_pfield(srvproc, SRV_SPID, (int *)NULL));
			break;
			
		case -1:
			// trace all connections
			server = SRV_GETSERVER(srvproc);
			config = SRV_GETCONFIG(server);
			srv_config(config, SRV_UNPACK_RPC,"TRUE",SRV_NULLTERM);
			srv_pre_handle(server,NULL,SRV_LANGUAGE,trace_lang,FALSE);
			srv_pre_handle(server,NULL,SRV_RPC,trace_rpc,FALSE);
			printf("%s  starting trace for all connections\n", _strtime(timestr));
			break;
		
		default:
			// Send error message and return
        	//
        	srv_sendmsg(srvproc, SRV_MSG_ERROR, TRACE_ERROR, SRV_INFO, (DBTINYINT)0,
		    NULL, 0, 0, "Error executing extended stored procedure: param not implemented",
                    SRV_NULLTERM);
	     	srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_MORE), 0, 0);
			return(XP_ERROR);			
		}
			    
    return(XP_NOERROR);

}

//
// XP_TRACEOFF
//	Used to remove the pre-handlers installed with XP_TRACEON
//	The first parameter should be the spid of the connection to be traced.
//	If 0, will trace the current spid.  If -1, trace all spids.
//
// Parameters:
//    srvproc - the handle to the client connection that got the SRV_CONNECT.
//
// Returns:
//    XP_NOERROR
//    XP_ERROR
//
//

RETCODE xp_traceoff(srvproc)
SRV_PROC *srvproc;
{
    int paramnum;
	int paramtype;
	int trace_num;
	SRV_SERVER *server;
	SRV_CONFIG *config;
	char timestr[9];

    // Check number, type of parameters
    //
	
    // 
	// Fetch param value.
	if (paramnum = 1) 
		srv_bmove(srv_paramdata(srvproc, 1), &trace_num, (DBINT) 4);
	else
		trace_num = 0;
	switch(trace_num) {
		case 0:
			// trace only this current connection
			srv_pre_handle(NULL,srvproc,SRV_LANGUAGE,trace_lang,TRUE);
			srv_pre_handle(NULL,srvproc,SRV_RPC,trace_rpc,TRUE);
			printf("%s  KPID %s  stopping trace \n", _strtime(timestr),
    			srv_pfield(srvproc, SRV_SPID, (int *)NULL));
			break;
			
		case -1:
			// trace all connections
			server = SRV_GETSERVER(srvproc);
			config = SRV_GETCONFIG(server);
			srv_config(config, SRV_UNPACK_RPC,"FALSE",SRV_NULLTERM);
			srv_pre_handle(server,NULL,SRV_LANGUAGE,trace_lang,FALSE);
			srv_pre_handle(server,NULL,SRV_RPC,trace_rpc,FALSE);
			printf("%s  stopping trace for all connections\n", _strtime(timestr));
			break;
		
		default:
			// Send error message and return
        	//
        	srv_sendmsg(srvproc, SRV_MSG_ERROR, TRACE_ERROR, SRV_INFO, (DBTINYINT)0,
		    NULL, 0, 0, "Error executing extended stored procedure: param not implemented",
                    SRV_NULLTERM);
	     	srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_MORE), 0, 0);
			return(XP_ERROR);			
		}
			    
    return(XP_NOERROR);

}
//****************************************************************************
// trace_lang
//      This is the pre-language handler routine to print language requests to 
//		stdio
//
// Parameters:
//    srvproc - a pointer to SRV_PROC stucture
//
// Returns:
//    SRV_CONTINUE
//
// Side Effects:
//
//****************************************************************************
RETCODE trace_lang(srvproc)
SRV_PROC *srvproc;
{
    char timestr[9];
	DBCHAR *query;
	int query_len;

	query= srv_langptr(srvproc);
	query_len = srv_langlen(srvproc);

   	//print the request
    printf("%s  KPID %s lang:  %s\n", _strtime(timestr),
    	srv_pfield(srvproc, SRV_SPID, (int *)NULL),query);
	if (query[query_len-1] != '\n')
		printf("\n");
	
    return(SRV_CONTINUE);
}

//****************************************************************************
// trace_rpc
//      This is the pre-rpc handler routine to print language requests to 
//		stdio
//
// Parameters:
//    srvproc - a pointer to SRV_PROC stucture
//
// Returns:
//    SRV_CONTINUE
//
// Side Effects:
//
//****************************************************************************
RETCODE trace_rpc(srvproc)
SRV_PROC *srvproc;
{
    int i;                  // Index variable
    short params;
    DBCHAR complete_rpc_name[4 * (MAXNAME + 1)]; //database.owner.name;#
	DBCHAR *rpc_db;
	DBCHAR *rpc_owner;
	DBCHAR *rpc_name;
	int rpc_number;
	char rpc_number_char[17];
	DBCHAR *rpc_paramname;
	BYTE rpc_paramstatus;
	int rpc_paramtype;
	DBINT rpc_parammaxlen;
	DBINT rpc_paramlen;
	void *rpc_paramdata;
	char rpc_paramstr[256];
	char timestr[9];
    

    // Construct full RPC name 
    //
	memset (complete_rpc_name, 0, sizeof(complete_rpc_name));
	rpc_db = srv_rpcdb(srvproc, (int *)NULL);
	rpc_owner = srv_rpcowner(srvproc, (int *)NULL);
	rpc_name = srv_rpcname(srvproc, (int *)NULL);
	rpc_number = srv_rpcnumber(srvproc);
	
	if (rpc_db != NULL) {
		strcat(complete_rpc_name,rpc_db);
		strcat(complete_rpc_name,".");
	};
	if (rpc_owner != NULL) {
		strcat(complete_rpc_name,rpc_owner);
		strcat(complete_rpc_name,".");
	};
	strcat(complete_rpc_name,rpc_name);
	if (rpc_number > 0) {
    	strcat(complete_rpc_name,";");
		_itoa(rpc_number,rpc_number_char,10);
		strcat(complete_rpc_name, rpc_number_char);
	};
	
	//start printing the request

    printf("%s  KPID %s rpc:  %s", _strtime(timestr),
    	srv_pfield(srvproc, SRV_SPID, (int *)NULL),complete_rpc_name);
    
    // Set up any RPC parameters.
    //
    params = srv_rpcparams(srvproc);

    for (i = 1; i <= params; i++) {

		rpc_paramname = srv_paramname(srvproc, i, (int *)NULL);
		if (strlen(rpc_paramname) == 0) rpc_paramname = NULL;

        rpc_paramstatus = (BYTE)srv_paramstatus(srvproc, i);
        rpc_paramtype = srv_paramtype(srvproc, i);
        rpc_parammaxlen = srv_parammaxlen(srvproc, i);
		rpc_paramlen = srv_paramlen(srvproc, i);
        if (rpc_paramlen > 0){
        	rpc_paramdata = srv_paramdata(srvproc, i);      		
			srv_convert(srvproc,rpc_paramtype,rpc_paramdata,rpc_paramlen,
						SRVCHAR,rpc_paramstr,-1);
		}
		else 
			strcpy(rpc_paramstr,"<null>");
		printf("  P%d: %s",i,rpc_paramstr);
		
		
    }
	printf("<end>\n\n");

    return(SRV_CONTINUE);
}



