    //-SLIDE 4-------------------------------------------------------------
    //
    // Create a read/write file cached
    //

    hFile = CreateFile(
                "foobar.txt",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ,
                NULL,
                CREATE_ALWAYS,
                FILE_ATTRIBUTE_NORMAL,
                NULL
                );

    //-SLIDE 4-------------------------------------------------------------
    //
    // unbuffered asynch I/O
    //

    hFile = CreateFile(
                "foobar.txt",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                OPEN_ALWAYS,
                FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING,
                NULL
                );

    //-SLIDE 4-------------------------------------------------------------
    //
    // Raw unbuffered asynch I/O
    //

    hFile = CreateFile(
                "\\\\.\\d:",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                OPEN_ALWAYS,
                FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING,
                NULL
                );

    //-SLIDE 4-------------------------------------------------------------
    //
    // I/O completion techniques
    //

    //-SLIDE 4--------------------------------------------------------------------
    //
    // waiting on an event
    //

    b1 = ReadFile(hFile,&buff,1, &n, &ov);

    //
    // if read succesful, or error with I/O pending, then pick up result
    //

    if ( b1 || (!b1 && GetLastError() == ERROR_IO_PENDING)) {

        b2 = GetOverlappedResult(
                hFile,
                &ov,
                &n2,
                TRUE
                );

            if ( b2 ) {
                printf("GetOv ok n2 %d hFile %x\n",
                    n2,
                    hFile
                    );
                }
            }


    //-SLIDE 4--------------------------------------------------------------------
    //
    // Using a completion callback
    //

    typedef struct _aio {
        LIST_ENTRY Links;
        HANDLE ReadHandle;
        HANDLE WriteHandle;
        LPVOID Buffer;
        OVERLAPPED Overlapped;
    } AIO, *PAIO;

    VOID
    ReadIoComplete(
        DWORD dwErrorCode,
        DWORD dwNumberOfBytesTransfered,
        LPOVERLAPPED lpOverlapped
        )
    {
        PAIO paio;
        BOOL IoOperationStatus;

        //
        // If an I/O error occured, display the error and then exit
        //

        if ( dwErrorCode ) {
            printf("FATAL I/O Error %ld I/O Context %lx.%lx\n",
                dwErrorCode,
                lpOverlapped,
                lpOverlapped->hEvent
                );
            ExitProcess(dwErrorCode);
            }
        paio = (PAIO)CONTAINING_RECORD(lpOverlapped,AIO,Overlapped);

        IoOperationStatus = WriteFileEx(
                                paio->WriteHandle,
                                paio->Buffer,
                                dwNumberOfBytesTransfered,
                                &paio->Overlapped,
                                WriteIoComplete
                                );

        //
        // Test to see if I/O was queued successfully
        //

        if ( !IoOperationStatus ) {
            if ( GetLastError() != ERROR_HANDLE_EOF ) {
                printf("FATAL I/O Error %ld\n",
                    GetLastError()
                    );
                ExitProcess(1);
                }
            }

    }

    //
    // main read loop
    //

    for(;;){

        //
        // Do an alertable wait on the handle vector. Both objects
        // becoming signaled at the same time means there is an
        // I/O request in the queue, and the caller has exclusive
        // access to the queue
        //

        CompletionStatus = WaitForMultipleObjectsEx(
                                2,
                                HandleVector,
                                TRUE,
                                0xffffffff,
                                TRUE
                                );

        //
        // If the wait failed, error out
        //

        if ( CompletionStatus == 0xffffffff ) {
            printf("FATAL WAIT ERROR %ld\n",GetLastError());
            ExitProcess(1);
            }

        //
        // If an I/O completion occured, then wait for another
        // I/O request or I/O completion
        //

        if ( CompletionStatus != WAIT_IO_COMPLETION ) {

            //
            // Wait was satisfied. We now have exclusive ownership of the
            // I/O request queue, and there is something on the queue.
            // Note that to insert something on the queue, the inserter
            // gets the list lock (mutex), inserts an entry, signals the
            // list semaphore, and finally releases the list lock
            //

            paio = (PAIO)RemoveHeadList(&IoRequestList);

            ReleaseMutex(IoWorkerListLock);

            IoOperationStatus = ReadFileEx(
                                    paio->ReadHandle,
                                    paio->Buffer,
                                    paio->Overlapped.Internal,
                                    &paio->Overlapped,
                                    ReadIoComplete
                                    );

            //
            // Test to see if I/O was queued successfully
            //

            if ( !IoOperationStatus ) {
                if ( GetLastError() != ERROR_HANDLE_EOF ) {
                    printf("FATAL I/O Error %ld\n",
                        GetLastError()
                        );
                    ExitProcess(1);
                    }
                }

            //
            // The I/O was successfully queued. Go back into the alertable
            // wait waiting for I/O completion, or for more I/O requests
            //

            }
        }

    //-SLIDE 4--------------------------------------------------------------------
    //
    // Using I/O completion ports
    //

    hFile = CreateFile(
                FileName,
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                fRaw ? OPEN_EXISTING : OPEN_ALWAYS,
                FileFlags,
                NULL
                );
    if ( hFile == INVALID_HANDLE_VALUE ) {
        printf("Error opening file %s %d\n",FileName,GetLastError());
        return 99;
        }

    hPort = CreateIoCompletionPort(
                hFile,
                NULL,               // new completion port
                (DWORD)hFile,       // per file context is the file handle
                0
                );

    ZeroMemory(&ov,sizeof(ov));
    ov.Offset = 0;
    PendingIoCount = 0;

    //
    // Issue the writes
    //

    for (WriteCount = 0; WriteCount < NumberOfWrites; WriteCount++){
reissuewrite:
        b = WriteFile(hFile,Buffer,BufferSize,&n,&ov);
        if ( !b && GetLastError() != ERROR_IO_PENDING ) {

            //
            // we reached our limit on outstanding I/Os
            //

            if ( GetLastError() == ERROR_INVALID_USER_BUFFER ||
                 GetLastError() == ERROR_NOT_ENOUGH_QUOTA ||
                 GetLastError() == ERROR_NOT_ENOUGH_MEMORY ) {

                //
                // wait for an outstanding I/O to complete and then go again
                //
                b = GetQueuedCompletionStatus(
                        hPort,
                        &n2,        // number of bytes
                        &key,       // per file context
                        &ov2,       // per I/O context is the overlapped struct
                        (DWORD)-1
                        );
                if ( !b ) {
                    printf("Error picking up completion write %d\n",GetLastError());
                    return 99;
                    }

                PendingIoCount--;
                goto reissuewrite;
                }
            else {
                printf("Error in write %d (pending count = %d)\n",GetLastError(),PendingIoCount);
                return 99;
                }
            }
        PendingIoCount++;
        ov.Offset += BufferSize;
        }

    //
    // Pick up the I/O completion
    //

    for (WriteCount = 0; WriteCount < PendingIoCount; WriteCount++){
        b = GetQueuedCompletionStatus(
                hPort,
                &n2,
                &key,
                &ov2,
                (DWORD)-1
                );
        if ( !b ) {
            printf("Error picking up completion write %d\n",GetLastError());
            return 99;
            }
        }

    //-SLIDE 4--------------------------------------------------------------------
    //
    // Taking an associated hFile and using wait completion instead of
    // completion port completion
    //

    //
    // Set hEvent and set the low order tag bit
    //

    ov.hEvent = (HANDLE)((DWORD)hev | 1);
    b1 = ReadFile(hFile,&buff,1, &n, &ov);

    //
    // if read succesful, or error with I/O pending, then pick up result
    //

    if ( b1 || (!b1 && GetLastError() == ERROR_IO_PENDING)) {

        b2 = GetOverlappedResult(
                hFile,
                &ov,
                &n2,
                TRUE
                );

            if ( b2 ) {
                printf("GetOv ok n2 %d hFile %x\n",
                    n2,
                    hFile
                    );
                }
            }


    //-SLIDE 5--------------------------------------------------------------------
    //
    // Setting a thread's affinity
    //

    //
    // limit to processor 0
    //

    SetThreadAffinityMask (GetCurrentThread (), 1);

    //
    // limit to processor 0 or 1
    //

    SetThreadAffinityMask (GetCurrentThread (), 3);




    //-SLIDE 6--------------------------------------------------------------------
    //
    // Setting a processes priority class
    //

    //
    // Idle Class base is 4
    //

    SetPriorityClass(GetCurrentProcess(),IDLE_PRIORITY_CLASS);

    //
    // Normal Class base is 9, 8, or 7 depending on foreground/background,
    // and setting of tasking options in system control panel applet
    //

    SetPriorityClass(GetCurrentProcess(),NORMAL_PRIORITY_CLASS);

    //
    // High Class base is 13
    //

    SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS);

    //
    // Realtime Class base is 24
    //

    SetPriorityClass(GetCurrentProcess(),REALTIME_PRIORITY_CLASS);


    //-SLIDE 6--------------------------------------------------------------------
    //
    // Setting a threads priority within a class
    //

    //
    // Highest is base of class + 2 points
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_HIGHEST);

    //
    // Above Normal is base of class + 1 points
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_ABOVE_NORMAL);

    //
    // Normal is base of class
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_NORMAL);

    //
    // Below Normal is base of class - 1 point
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_BELOW_NORMAL);

    //
    // Lowest is base of class - 2 points
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_LOWEST);

    //
    // Idle is 1 for Dynamic (idle, normal, high) and 16 for Realtime
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_IDLE);

    //
    // Time critical is 15 for Dynamic (idle, normal, high) and 31 for Realtime
    //

    SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_TIME_CRITICAL);



    //-SLIDE 9--------------------------------------------------------------------
    //
    // Critical Section Object
    //

    //
    // Initialize a critical section before using it
    //

    InitializeCriticalSection(&CriticalSection);

    //
    // Use EnterCriticalSection to aquire ownership
    //

    EnterCriticalSection(&CriticalSection);


    //
    // When you are done in your single threaded region, use LeaveCriticalSection
    //

    LeaveCriticalSection(&CriticalSection);


    //
    // Use DeleteCriticalSection when you are done with the object
    //

    DeleteCriticalSection(&CriticalSection);


    //-SLIDE 10-------------------------------------------------------------------
    //
    // Event Objects

    //
    // Create a manual reset event whose initial state is false and is unnamed
    //

    IoReadsDoneEvent = CreateEvent(NULL,TRUE,FALSE,NULL);

    //
    // One thread might signal that the event occured by setting the event
    //

    SetEvent(IoReadsDoneEvent);

    //
    // Another thread might wait for the event to occur by waiting on it
    //

    WaitForSingleObject(IoReadsDoneEvent,-1);

    //
    // Sometime you need to wait for more than one event to be set at the same time
    // use WaitForMultipleObjects in this scenario
    //
    // Wait for both events to be signaled at the same time (bWaitAll == TRUE)
    //

    CompletionHandles[0] = IoReadsDoneEvent;
    CompletionHandles[1] = IoWritesDoneEvent;
    WaitStatus = WaitForMultipleObjects(2,CompletionHandles,TRUE,0xffffffff);

    //
    // Use CloseHandle to release resources of the event
    //

    CloseHandle(IoReadsDoneEvent);


    //-SLIDE 11--------------------------------------------------------------------
    //
    // Mutex Objects

    //
    // create an unnamed mutex that is NOT initially owned by the caller
    //

    Mutex = CreateMutex(NULL,FALSE,NULL);

    //
    // Aquire ownership of the mutex using wait
    //

    WaitStatus = WaitForSingleObject(Mutex,-1);

    if ( WaitStatus == WAIT_ABANDONED ) {

        //
        // I have ownership, but the last owner died while holding
        // this lock so attempt cleanup
        //

        cleanup();

        }

    //
    // When you are done in your single threaded region, use ReleaseMutex
    //

    ReleaseMutex(Mutex);

    //
    // Use CloseHandle to release resources of the mutex
    //

    CloseHandle(Mutex);


    //-SLIDE 12--------------------------------------------------------------------
    //
    // Implementing a synchronized queue using a mutex and a semaphore

    //
    // Initialize a mutex and a semaphore
    //

    IoWorkerListLock = CreateMutex(NULL,FALSE,NULL);;
    IoWorkerListSemaphore = CreateSemaphore(NULL,0,0x7fffffff,NULL);
    InitializeListHead(&IoRequestList);

    //
    // To put an item on the list, grab the list mutex, insert your entry,
    // signal the semaphore, and release the mutex
    //

    WaitStatus = WaitForSingleObject(IoWorkerListLock,-1);
    InsertTailList(&IoRequestList,&paio->Links);
    ReleaseSemaphore(IoWorkerListSemaphore,1,NULL);
    ReleaseMutex(IoWorkerListLock);

    //
    // A thread can remove something from the list by waiting on the mutex
    // AND the semaphore. When both of these objects are signalled, the
    // thread has ownership of the list, and causes the semephore count
    // to decrement to account for the item being removed from the list
    //

    HandleVector[0] = IoWorkerListLock;
    HandleVector[1] = IoWorkerListSemaphore;

    WaitForMultipleObjects(
        2,
        HandleVector,
        TRUE,
        0xffffffff
        );

    //
    // thread has ownership of the list head, so remove an entry and
    // then release the list head lock
    //

    paio = (PAIO)RemoveHeadList(&IoRequestList);
    ReleaseMutex(IoWorkerListLock);

    //
    // Use CloseHandle to release resources of the semaphore
    //

    CloseHandle(IoWorkerListSemaphore);


    //-SLIDE 13--------------------------------------------------------------------
    //
    // Global Critical Section object
    //

#include <stdio.h>
#include <windows.h>


//
// Global Critical Sections have two components. One piece is shared between all
// applications using the global lock. This portion will typically reside in some
// sort of shared memory
//
// The second piece is per-process. This contains a per-process handle to the shared
// critical section lock semaphore. The semaphore is itself shared, but each process
// may have a different handle value to the semaphore.
//
// Global critical sections are attached to by name. The application wishing to
// attach must know the name of the critical section (actually the name of the shared
// lock semaphore, and must know the address of the global portion of the critical
// section
//

typedef struct _GLOBAL_SHARED_CRITICAL_SECTION {
    LONG LockCount;
    LONG RecursionCount;
    DWORD OwningThread;
    DWORD Reserved;
} GLOBAL_SHARED_CRITICAL_SECTION, *PGLOBAL_SHARED_CRITICAL_SECTION;

typedef struct _GLOBAL_LOCAL_CRITICAL_SECTION {
    PGLOBAL_SHARED_CRITICAL_SECTION GlobalPortion;
    HANDLE LockSemaphore;
    DWORD Reserved1;
    DWORD Reserved2;
} GLOBAL_LOCAL_CRITICAL_SECTION, *PGLOBAL_LOCAL_CRITICAL_SECTION;


BOOL
WINAPI
AttachToGlobalCriticalSection(
    PGLOBAL_LOCAL_CRITICAL_SECTION lpLocalPortion,
    PGLOBAL_SHARED_CRITICAL_SECTION lpGlobalPortion,
    LPCSTR lpName
    )

/*++

Routine Description:

    This routine attaches to an existing global critical section, or creates and
    initializes the global critical section if it does not already exist.

Arguments:

    lpLocalPortion - Supplies the address of a per-app local portion of the global
        critical section.

    lpGlobalPortion - Supplies the address of the global shared portion of the
        critical section. If the critical section is new, the caller will initialize it.

    lpName - Supplies the name of the critical section.  If an existing
        critical section with this name already exists, then it is not
        reinitialized.  In this case, the caller simply attaches to it.

Return Value:

    TRUE - The operation was successful.

    FALSE - The operation failed.

--*/

{

    HANDLE GlobalMutex;
    HANDLE LockSemaphore;
    BOOL rv;
    DWORD WaitResult;

    //
    // Serialize all global critical section initialization
    //

    GlobalMutex = CreateMutex(NULL,TRUE,"GlobalCsMutex");

    //
    // If the mutex create/open failed, then bail
    //

    if ( !GlobalMutex ) {
        return FALSE;
        }

    if ( GetLastError() == ERROR_ALREADY_EXISTS ) {

        //
        // Since the mutex already existed, the request for ownership has no effect.
        // wait for the mutex
        //

        WaitResult = WaitForSingleObject(GlobalMutex,INFINITE);
        if ( WaitResult == WAIT_FAILED ) {
            CloseHandle(GlobalMutex);
            return FALSE;
            }
        }

    //
    // We now own the global critical section creation mutex. Create/Open the
    // named semaphore. If we are the creator, then initialize the critical
    // section. Otherwise just point to it. The global critical section creation
    // allows us to do this safely.
    //

    rv = FALSE;
    LockSemaphore = NULL;
    try {
        LockSemaphore = CreateSemaphore(NULL,0,MAXLONG-1,lpName);

        //
        // If the semaphore create/open failed, then bail
        //

        if ( !GlobalMutex ) {
            rv = FALSE;
            goto finallyexit;
            }

        //
        // See if we attached to the semaphore, or if we created it. If we created it,
        // then we need to init the global structure.
        //

        if ( GetLastError() != ERROR_ALREADY_EXISTS ) {

            //
            // We Created the semaphore, so init the global portion.
            //

            lpGlobalPortion->LockCount = -1;
            lpGlobalPortion->RecursionCount = 0;
            lpGlobalPortion->OwningThread = 0;
            lpGlobalPortion->Reserved = 0;
            }

        lpLocalPortion->LockSemaphore = LockSemaphore;
        LockSemaphore = NULL;
        lpLocalPortion->GlobalPortion = lpGlobalPortion;
        lpLocalPortion->Reserved1 = 0;
        lpLocalPortion->Reserved2 = 0;
        rv = TRUE;
finallyexit:;
        }
    finally {
        ReleaseMutex(GlobalMutex);
        CloseHandle(GlobalMutex);
        if ( LockSemaphore ) {
            CloseHandle(LockSemaphore);
            }
        }

    return rv;
}

BOOL
WINAPI
DetachFromGlobalCriticalSection(
    PGLOBAL_LOCAL_CRITICAL_SECTION lpLocalPortion
    )

/*++

Routine Description:

    This routine detaches from an existing global critical section.

Arguments:

    lpLocalPortion - Supplies the address of a per-app local portion of the global
        critical section.

Return Value:

    TRUE - The operation was successful.

    FALSE - The operation failed.

--*/

{

    HANDLE LockSemaphore;
    HANDLE GlobalMutex;
    DWORD WaitResult;
    BOOL rv;


    //
    // Serialize all global critical section initialization
    //

    GlobalMutex = CreateMutex(NULL,TRUE,"GlobalCsMutex");

    //
    // If the mutex create/open failed, then bail
    //

    if ( !GlobalMutex ) {
        return FALSE;
        }

    if ( GetLastError() == ERROR_ALREADY_EXISTS ) {

        //
        // Since the mutex already existed, the request for ownership has no effect.
        // wait for the mutex
        //

        WaitResult = WaitForSingleObject(GlobalMutex,INFINITE);
        if ( WaitResult == WAIT_FAILED ) {
            CloseHandle(GlobalMutex);
            return FALSE;
            }
        }
    LockSemaphore = NULL;
    rv = FALSE;
    try {
        LockSemaphore = lpLocalPortion->LockSemaphore;
        ZeroMemory(lpLocalPortion,sizeof(*lpLocalPortion));
        rv = TRUE;
        }
    finally {
        if ( LockSemaphore ) {
            CloseHandle(LockSemaphore);
            }
        ReleaseMutex(GlobalMutex);
        CloseHandle(GlobalMutex);
        }
    return rv;
}

VOID
WINAPI
EnterGlobalCriticalSection(
    PGLOBAL_LOCAL_CRITICAL_SECTION lpLocalPortion
    )
{
    PGLOBAL_SHARED_CRITICAL_SECTION GlobalPortion;
    DWORD ThreadId;
    LONG IncResult;
    DWORD WaitResult;

    ThreadId = GetCurrentThreadId();
    GlobalPortion = lpLocalPortion->GlobalPortion;

    //
    // Increment the lock variable. On the transition to 0, the caller
    // becomes the absolute owner of the lock. Otherwise, the caller is
    // either recursing, or is going to have to wait
    //

    IncResult = InterlockedIncrement(&GlobalPortion->LockCount);
    if ( !IncResult ) {

        //
        // lock count went from 0 to 1, so the caller
        // is the owner of the lock
        //

        GlobalPortion->RecursionCount = 1;
        GlobalPortion->OwningThread = ThreadId;
        }
    else {

        //
        // If the caller is recursing, then increment the recursion count
        //

        if ( GlobalPortion->OwningThread == ThreadId ) {
            GlobalPortion->RecursionCount++;
            }
        else {
            WaitResult = WaitForSingleObject(lpLocalPortion->LockSemaphore,INFINITE);
            if ( WaitResult == WAIT_FAILED ) {
                RaiseException(GetLastError(),0,0,NULL);
                }
            GlobalPortion->RecursionCount = 1;
            GlobalPortion->OwningThread = ThreadId;
            }
        }
}

VOID
WINAPI
LeaveGlobalCriticalSection(
    PGLOBAL_LOCAL_CRITICAL_SECTION lpLocalPortion
    )
{
    PGLOBAL_SHARED_CRITICAL_SECTION GlobalPortion;
    LONG DecResult;

    GlobalPortion = lpLocalPortion->GlobalPortion;


    //
    // decrement the recursion count. If it is still non-zero, then
    // we are still the owner so don't do anything other than dec the lock
    // count
    //

    if (--GlobalPortion->RecursionCount) {
        InterlockedDecrement(&GlobalPortion->LockCount);
        }
    else {

        //
        // We are really leaving, so give up ownership and decrement the
        // lock count
        //

        GlobalPortion->OwningThread = 0;
        DecResult = InterlockedDecrement(&GlobalPortion->LockCount);

        //
        // Check to see if there are other waiters. If so, then wake up a waiter
        //

        if ( DecResult >= 0 ) {
            ReleaseSemaphore(lpLocalPortion->LockSemaphore,1,NULL);
            }

        }
}

GLOBAL_LOCAL_CRITICAL_SECTION LocalPortion;

int _CRTAPI1
main(
    int argc,
    char *argv[],
    char *envp[]
    )
{

    HANDLE hFileMap;
    LPVOID SharedMem;
    BOOL b;
    int i;
    DWORD Start,End;
    HANDLE Mutex1;

    //
    // open or create a shared file mapping object
    //

    hFileMap = CreateFileMapping((HANDLE)0xffffffff,NULL,PAGE_READWRITE,0,1024,"MyMem");

    if ( !hFileMap ) {
        printf("create file map failed\n");
        ExitProcess(1);
        }

    SharedMem = MapViewOfFile(hFileMap,FILE_MAP_WRITE,0,0,0);

    if ( !SharedMem ) {
        printf("map view failed\n");
        ExitProcess(1);
        }

    b = AttachToGlobalCriticalSection(&LocalPortion,SharedMem,"MyGlobalCs");

    if ( !b ) {
        printf("attach failed\n");
        ExitProcess(1);
        }

    if ( argc > 1 ) {

        for(i=0;i<30;i++){
            EnterGlobalCriticalSection(&LocalPortion);
            printf("Thread %x is in\n",GetCurrentThreadId());
            Sleep(500);
            LeaveGlobalCriticalSection(&LocalPortion);
            }
        }

    Start = GetTickCount();
    for(i=0;i<1000000;i++){
        EnterGlobalCriticalSection(&LocalPortion);
        LeaveGlobalCriticalSection(&LocalPortion);
        }
    End = GetTickCount();
    printf("Global CS Time %dms\n",End-Start);

    Mutex1 = CreateMutex(NULL,FALSE,NULL);
    Start = GetTickCount();
    for(i=0;i<100000;i++){
        WaitForSingleObject(Mutex1,INFINITE);
        ReleaseMutex(Mutex1);
        }
    End = GetTickCount();
    printf("Mutex Time     %dms\n",(End-Start)*10);

    DetachFromGlobalCriticalSection(&LocalPortion);
}


    //-SLIDE 14--------------------------------------------------------------------
    //
    // How to call SetProcessWorkingSetSize
    //

    //
    // To trim your working set (manual swap). VC++ does this in the GUI
    // in preperation for a windowless compile
    //

    SetProcessWorkingSetSize(GetCurrentProcess(),-1,-1);

    //
    // To set your minimum working set to 1mb and max to 4mb use
    // This is usefull if you use heavy cached I/O and have a predictable
    // internal memory load in your server. If you know your working set,
    // you can minimize paging in your app
    //

    SetProcessWorkingSetSize(GetCurrentProcess(),1048576,4194304);


    //-SLIDE 14--------------------------------------------------------------------
    //
    // How to call VirtualLock
    //
    // Lock 4096 bytes at location ReadWrite2
    //

    VirtualLock(ReadWrite2,4096);


    //-SLIDE 14--------------------------------------------------------------------
    //
    // How to use memory mapped files
    //

    //
    // To map a real file ( see global critical section example for named
    // shared memory example
    //
    //

    //
    // Open a handle to a file
    //

    FileHandle = CreateFile(
                    ImageName,
                    GENERIC_READ | GENERIC_WRITE,
                    FILE_SHARE_READ,
                    NULL,
                    OPEN_EXISTING,
                    0,
                    NULL
                    );

    //
    // Create a file mapping object for the entire file
    //

    MappingHandle = CreateFileMapping(
                        FileHandle,
                        NULL,
                        PAGE_READWRITE,
                        0,
                        0,
                        NULL
                        );
    //
    // Release our handle to the file
    //

    CloseHandle( FileHandle );


    //
    // Map the file into our address space
    //

    BaseAddress = MapViewOfFile(
                        MappingHandle,
                        FILE_MAP_READ | FILE_MAP_WRITE,
                        0,
                        0,
                        0
                        );
    //
    // Data in the file is now accessable using pointers based at BaseAddress
    //

    //
    // Release our handle to the mapping object
    //

    CloseHandle( MappingHandle );

    //
    // Unmap the file when you are done
    //

    UnmapViewOfFile( BaseAddress );
