    //-SLIDE 4-------------------------------------------------------------
    //
    // Create a read/write file cached
    //

    hFile = CreateFile(
                "foobar.txt",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ,
                NULL,
                CREATE_ALWAYS,
                FILE_ATTRIBUTE_NORMAL,
                NULL
                );

    //-SLIDE 4-------------------------------------------------------------
    //
    // unbuffered asynch I/O
    //

    hFile = CreateFile(
                "foobar.txt",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                OPEN_ALWAYS,
                FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING,
                NULL
                );

    //-SLIDE 4-------------------------------------------------------------
    //
    // Raw unbuffered asynch I/O
    //

    hFile = CreateFile(
                "\\\\.\\d:",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                OPEN_ALWAYS,
                FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING,
                NULL
                );

    //-SLIDE 4-------------------------------------------------------------
    //
    // I/O completion techniques
    //

    //-SLIDE 4--------------------------------------------------------------------
    //
    // waiting on an event
    //

    b1 = ReadFile(hFile,&buff,1, &n, &ov);

    //
    // if read succesful, or error with I/O pending, then pick up result
    //

    if ( b1 || (!b1 && GetLastError() == ERROR_IO_PENDING)) {

        b2 = GetOverlappedResult(
                hFile,
                &ov,
                &n2,
                TRUE
                );

            if ( b2 ) {
                printf("GetOv ok n2 %d hFile %x\n",
                    n2,
                    hFile
                    );
                }
            }


    //-SLIDE 4--------------------------------------------------------------------
    //
    // Using a completion callback
    //

    typedef struct _aio {
        LIST_ENTRY Links;
        HANDLE ReadHandle;
        HANDLE WriteHandle;
        LPVOID Buffer;
        OVERLAPPED Overlapped;
    } AIO, *PAIO;

    VOID
    ReadIoComplete(
        DWORD dwErrorCode,
        DWORD dwNumberOfBytesTransfered,
        LPOVERLAPPED lpOverlapped
        )
    {
        PAIO paio;
        BOOL IoOperationStatus;

        //
        // If an I/O error occured, display the error and then exit
        //

        if ( dwErrorCode ) {
            printf("FATAL I/O Error %ld I/O Context %lx.%lx\n",
                dwErrorCode,
                lpOverlapped,
                lpOverlapped->hEvent
                );
            ExitProcess(dwErrorCode);
            }
        paio = (PAIO)CONTAINING_RECORD(lpOverlapped,AIO,Overlapped);

        IoOperationStatus = WriteFileEx(
                                paio->WriteHandle,
                                paio->Buffer,
                                dwNumberOfBytesTransfered,
                                &paio->Overlapped,
                                WriteIoComplete
                                );

        //
        // Test to see if I/O was queued successfully
        //

        if ( !IoOperationStatus ) {
            if ( GetLastError() != ERROR_HANDLE_EOF ) {
                printf("FATAL I/O Error %ld\n",
                    GetLastError()
                    );
                ExitProcess(1);
                }
            }

    }

    //
    // main read loop
    //

    for(;;){

        //
        // Do an alertable wait on the handle vector. Both objects
        // becoming signaled at the same time means there is an
        // I/O request in the queue, and the caller has exclusive
        // access to the queue
        //

        CompletionStatus = WaitForMultipleObjectsEx(
                                2,
                                HandleVector,
                                TRUE,
                                0xffffffff,
                                TRUE
                                );

        //
        // If the wait failed, error out
        //

        if ( CompletionStatus == 0xffffffff ) {
            printf("FATAL WAIT ERROR %ld\n",GetLastError());
            ExitProcess(1);
            }

        //
        // If an I/O completion occured, then wait for another
        // I/O request or I/O completion
        //

        if ( CompletionStatus != WAIT_IO_COMPLETION ) {

            //
            // Wait was satisfied. We now have exclusive ownership of the
            // I/O request queue, and there is something on the queue.
            // Note that to insert something on the queue, the inserter
            // gets the list lock (mutex), inserts an entry, signals the
            // list semaphore, and finally releases the list lock
            //

            paio = (PAIO)RemoveHeadList(&IoRequestList);

            ReleaseMutex(IoWorkerListLock);

            IoOperationStatus = ReadFileEx(
                                    paio->ReadHandle,
                                    paio->Buffer,
                                    paio->Overlapped.Internal,
                                    &paio->Overlapped,
                                    ReadIoComplete
                                    );

            //
            // Test to see if I/O was queued successfully
            //

            if ( !IoOperationStatus ) {
                if ( GetLastError() != ERROR_HANDLE_EOF ) {
                    printf("FATAL I/O Error %ld\n",
                        GetLastError()
                        );
                    ExitProcess(1);
                    }
                }

            //
            // The I/O was successfully queued. Go back into the alertable
            // wait waiting for I/O completion, or for more I/O requests
            //

            }
        }

    //-SLIDE 4--------------------------------------------------------------------
    //
    // Using I/O completion ports
    //

    hFile = CreateFile(
                FileName,
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                fRaw ? OPEN_EXISTING : OPEN_ALWAYS,
                FileFlags,
                NULL
                );
    if ( hFile == INVALID_HANDLE_VALUE ) {
        printf("Error opening file %s %d\n",FileName,GetLastError());
        return 99;
        }

    hPort = CreateIoCompletionPort(
                hFile,
                NULL,               // new completion port
                (DWORD)hFile,       // per file context is the file handle
                0
                );

    ZeroMemory(&ov,sizeof(ov));
    ov.Offset = 0;
    PendingIoCount = 0;

    //
    // Issue the writes
    //

    for (WriteCount = 0; WriteCount < NumberOfWrites; WriteCount++){
reissuewrite:
        b = WriteFile(hFile,Buffer,BufferSize,&n,&ov);
        if ( !b && GetLastError() != ERROR_IO_PENDING ) {

            //
            // we reached our limit on outstanding I/Os
            //

            if ( GetLastError() == ERROR_INVALID_USER_BUFFER ||
                 GetLastError() == ERROR_NOT_ENOUGH_QUOTA ||
                 GetLastError() == ERROR_NOT_ENOUGH_MEMORY ) {

                //
                // wait for an outstanding I/O to complete and then go again
                //
                b = GetQueuedCompletionStatus(
                        hPort,
                        &n2,        // number of bytes
                        &key,       // per file context
                        &ov2,       // per I/O context is the overlapped struct
                        (DWORD)-1
                        );
                if ( !b ) {
                    printf("Error picking up completion write %d\n",GetLastError());
                    return 99;
                    }

                PendingIoCount--;
                goto reissuewrite;
                }
            else {
                printf("Error in write %d (pending count = %d)\n",GetLastError(),PendingIoCount);
                return 99;
                }
            }
        PendingIoCount++;
        ov.Offset += BufferSize;
        }

    //
    // Pick up the I/O completion
    //

    for (WriteCount = 0; WriteCount < PendingIoCount; WriteCount++){
        b = GetQueuedCompletionStatus(
                hPort,
                &n2,
                &key,
                &ov2,
                (DWORD)-1
                );
        if ( !b ) {
            printf("Error picking up completion write %d\n",GetLastError());
            return 99;
            }
        }

    //-SLIDE 4--------------------------------------------------------------------
    //
    // Taking an associated hFile and using wait completion instead of
    // completion port completion
    //

    //
    // Set hEvent and set the low order tag bit
    //

    ov.hEvent = (HANDLE)((DWORD)hev | 1);
    b1 = ReadFile(hFile,&buff,1, &n, &ov);

    //
    // if read succesful, or error with I/O pending, then pick up result
    //

    if ( b1 || (!b1 && GetLastError() == ERROR_IO_PENDING)) {

        b2 = GetOverlappedResult(
                hFile,
                &ov,
                &n2,
                TRUE
                );

            if ( b2 ) {
                printf("GetOv ok n2 %d hFile %x\n",
                    n2,
                    hFile
                    );
                }
            }



