/*
	Example: Service Controller API

	Start a service if not already started.
*/

VOID StartSampleService(LPCTSTR lpszService, 
                        DWORD dwNumArgs,
                        LPCTSTR *lpszArgs )
{
  SERVICE_STATUS ssStatus;
  DWORD dwOldCheckPoint;
  SC_HANDLE schSCManager;
  SC_HANDLE schService;

  schSCManager = OpenSCManager(
       NULL,                   /* local machine              */
       NULL,                   /* ServicesActive database    */
       SC_MANAGER_CONNECT);    /* Access to start a service  */

  if (schSCManager == NULL)
      MyErrorExit("OpenSCManager Failed");

  schService = OpenService(
       schSCManager,          /* S. C. Manager database */
       lpszService,           /* name of service        */
       SERVICE_START | SERVICE_QUERY_STATUS);

  CloseServiceHandle(schSCManager);

  if (schService == NULL)
      MyErrorExit("OpenService Failed");

  if (! StartService(schService, /* handle of service   */
            dwNumArgs,                   /* number of arguments */
            lpszArgs) )                  /* arguments        */
  {
      CloseServiceHandle(schService);
      if (GetLastError() == ERROR_SERVICE_ALREADY_RUNNING)
      {
          printf("Service already running");
          return;
      }
      MyErrorExit("StartService Failed");
  }
  else
      printf("Service start pending\n");

  /* Check the status until the service is running. */

  if (! QueryServiceStatus(
      schService,   /* handle of service       */
      &ssStatus) )  /* address of status info  */
  {
      CloseServiceHandle(schService);
      MyErrorExit("QueryServiceStatus Failed");
  }

  while (ssStatus.dwCurrentState != SERVICE_RUNNING)
  {
      /* Save the current checkpoint. */

      dwOldCheckPoint = ssStatus.dwCheckPoint;

      /* Wait for the specified interval. */


      Sleep(ssStatus.dwWaitHint);

      /* Check the status again. */

      if (! QueryServiceStatus(
              schService,   /* handle of service       */
              &ssStatus) )  /* address of status info  */
          break;

      /* Break if the checkpoint has not been incremented. */

      if (dwOldCheckPoint >= ssStatus.dwCheckPoint)
          break;

  }

  if (ssStatus.dwCurrentState == SERVICE_RUNNING)
      printf("StartService SUCCESS\n");
  else
  {
      printf("\nService not started: \n");

      printf("  Current State: %d\n",
                   ssStatus.dwCurrentState);
      printf("  Exit Code: %d\n", ssStatus.dwWin32ExitCode);
      printf("  Service Specific Exit Code: %d\n",
                   ssStatus.dwServiceSpecificExitCode);
      printf("  Check Point: %d\n", ssStatus.dwCheckPoint);
      printf("  Wait Hint: %d\n", ssStatus.dwWaitHint);
  }


  CloseServiceHandle(schService);

}
